/*
    This source file is part of the trafikinfo for android project
    Copyright (C) 2010  Peter Björkman


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package se.peterbjorkman.android.trafikinfo;

import com.google.android.maps.GeoPoint;
import com.google.android.maps.MyLocationOverlay;

import android.os.Bundle;
import android.util.Log;

public abstract class AutoZoomSituationMapActivity extends SituationMapActivity {
    private static final String LOGTAG = AutoZoomSituationMapActivity.class.getName();
    private static final int MIN_LAT_SPAN = 10000; // TODO hardcoded maxzoom
    private static final float ZOOM_MARGIN = 1.15F;
    private static final String INSTANCE_STATE_OWNLOCATIONRECIEVED = "OwnLocationRecieved"; // TODO hardcoded maxzoom
    private Boolean mOwnLocationRecieved = false;
    
    @Override
    public void onCreate(Bundle savedInstanceState) {
        Log.d(LOGTAG, "In onCreate");
        super.onCreate(savedInstanceState);
        if (savedInstanceState == null)
            zoomAndPan(false);
        else
            mOwnLocationRecieved = savedInstanceState.getBoolean(INSTANCE_STATE_OWNLOCATIONRECIEVED);
     }
    
    @Override
    protected void onFixedLocation() {
        if (mOwnLocationRecieved)
            return; // do not zoomAndPan if we have rotated the phone
        zoomAndPan(true);
        mOwnLocationRecieved = true;
    }
    
    @Override
    protected void onSaveInstanceState (Bundle outState){
        Log.d(LOGTAG, "In onCreate");
        super.onSaveInstanceState(outState);
        outState.putBoolean(INSTANCE_STATE_OWNLOCATIONRECIEVED, mOwnLocationRecieved);
    }

    private void zoomAndPan(Boolean lookAtMyLocation) {
        Limits itemsLimit = getLimitsFromItemizedOverlay();
        
        Limits limits;
        if (lookAtMyLocation){
            Limits locationLimits = getLimitsFromMyLocationOverlay();
            limits = new Limits(locationLimits, itemsLimit);
        } else {
            limits = itemsLimit;
        }
        
        setCenter(limits.getCenter());
     
        int latSpan = limits.getLatSpan();
        if (latSpan < MIN_LAT_SPAN)
            latSpan = MIN_LAT_SPAN;
        
        int longSpan = limits.getLongSpan();
        zoomToSpan((int)(latSpan * ZOOM_MARGIN), (int)(longSpan * ZOOM_MARGIN));
    }
    
    private Limits getLimitsFromItemizedOverlay(){
        SituationMapOverlay overlay = getItemizedoverlay();
        
        int latSpan = overlay.getLatSpanE6();
        int longSpan = overlay.getLonSpanE6();
        GeoPoint point = overlay.getCenter();
        
        if (point == null)
            return new Limits();
        
        return new Limits(point, latSpan, longSpan);
    }
    
    private Limits getLimitsFromMyLocationOverlay(){
        MyLocationOverlay overlay = getMyLocationOverlay();
        //float accurency = overlay.getLastFix().getAccuracy();

        GeoPoint point = overlay.getMyLocation();

        if (point == null)
            return new Limits();

        int latSpan = 1000; // TODO calculate span from accurency and location
        int longSpan = 1000;

        return new Limits(point, latSpan, longSpan);
    }
    
    private class Limits{

        public Limits(){
            hasLimits = false;
        }
        
        public Limits(GeoPoint point, int latSpan, int longSpan){
            int centerLong = point.getLongitudeE6();
            int centerLat  = point.getLatitudeE6();
            
            maxLat  = centerLat  + latSpan  / 2;
            minLat  = centerLat  - latSpan  / 2;
            maxLong = centerLong + longSpan / 2;
            minLong = centerLong - longSpan / 2;
            hasLimits = true;
        }
        
        public Limits(Limits l1, Limits l2){
            if (l1.hasLimits && l2.hasLimits){
                maxLong = Math.max(l1.getMaxLong(), l2.getMaxLong());
                maxLat  = Math.max(l1.getMaxLat() , l2.getMaxLat());
                minLong = Math.min(l1.getMinLong(), l2.getMinLong());
                minLat  = Math.min(l1.getMinLat(), l2.getMinLat());
                hasLimits = true;
            } else if (l1.hasLimits()){
                maxLong = l1.getMaxLong();
                maxLat  = l1.getMaxLat();
                minLong = l1.getMinLong();
                minLat  = l1.getMinLat();
                hasLimits = true;
            } else if (l2.hasLimits()){
                maxLong = l1.getMaxLong();
                maxLat  = l1.getMaxLat();
                minLong = l1.getMinLong();
                minLat  = l1.getMinLat();
                hasLimits = true;
            } else {
                hasLimits = false;
            }        
        }
        
        private int maxLong;
        private int maxLat;
        private int minLong;
        private int minLat;
        private Boolean hasLimits;
        
        public int getMaxLong(){
            return maxLong;
        }
        
        public int getMaxLat(){
            return maxLat;
        }
        
        public int getMinLong(){
            return minLong;
        }
        
        public int getMinLat(){
            return minLat;
        }
        
        public Boolean hasLimits(){
            return hasLimits;
        }
        
        private int average(int x1, int x2){
            return (x1 + x2) / 2;
        }
        
        public GeoPoint getCenter(){
            return new GeoPoint(average(maxLat, minLat), average(maxLong, minLong));
        }

        public int getLatSpan(){
            return maxLat - minLat;
        }
        
        public int getLongSpan(){
            return maxLong - minLong;
        }
    }
}
