/*
    This source file is part of the trafikinfo for android project
    Copyright (C) 2010  Peter Björkman


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package se.peterbjorkman.android.trafikinfo;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import se.peterbjorkman.android.trafikinfo.TrafficInfo.ReadState;

import android.content.ContentValues;
import android.content.Context;
import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;

public class InfoDbAdapter {
    private static final String LOGTAG = Logger.createLogtag(InfoDbAdapter.class);
    private DatabaseHelper mDbHelper;
    private SQLiteDatabase mDb;

    public static final String KEY_ID = "_id";
    public static final String KEY_EXTERNALID = "externalid";
    public static final String KEY_HEADER = "header";
    public static final String KEY_TYPE = "type";
    public static final String KEY_TYPENAME = "typename";
    public static final String KEY_SOURCE = "source";
    public static final String KEY_MODIFIED = "modified";
    public static final String KEY_LOCATION = "location";
    public static final String KEY_ROAD = "road";
    public static final String KEY_LONGITUDE = "longitude";
    public static final String KEY_LATITUDE = "latitude";
    public static final String KEY_FROM = "fromTime";
    public static final String KEY_TO = "toTime";
    public static final String KEY_OBSOLETE = "obsoleteTime";
    public static final String KEY_READ = "read";
    public static final String KEY_CATEGORY = "category";    

    public static final String KEY_TIMESTAMP = "timestamp";
    
    public static final String KEY_NUMBER = "number";
    public static final String KEY_TEXT = "text";
    public static final String KEY_INFO_ID = "info_id";
    
    public static final String KEY_SPEED_LIMIT = "speedlimit";
    
    private static final String KEY_RESTRICTION_NUMBER = "restrictionNumber";
    
    private static final String DATABASE_NAME = "data";
    private static final String INFORMATION_TABLE = "info"; // if renaming this table be sure to drop the old one in the upgrade script
    private static final String MESSAGE_TABLE = "message"; // TODO rename comment table 
    private static final String TIMESTAMP_TABLE = "timestamp";
    private static final String RESTRICTION_TABLE = "restriction";
    private static final String RESTRICTION_COMMENT_TABLE = "restrictionComment";
    private static final String POSITION_TABLE = "position";
    
    private static final int DATABASE_VERSION = 5;
    
    private static final int TIMESTAMPTYPE_LASTUPDATE = 1;
    private static final int TIMESTAMPTYPE_SERVERTIME = 2;
    private static final int TIMESTAMPTYPE_LASTUPATECOUNTY = 3;
    private static final int TIMESTAMPTYPE_LASTVERSION = 4;
    private static final int TIMESTAMPTYPE_LASTCLEANUPTIME = 5;
    
    public static final int POSITIONTYPE_LASTPOSITION = 1;
    public static final int POSITIONTYPE_LASTFETCHMIN = 2;
    public static final int POSITIONTYPE_LASTFETCHMAX = 3;

    private static final String REVERSE_ORDERBY = KEY_READ + ", " + KEY_FROM;

    private static final String ORDERBY = KEY_READ + ", " + KEY_FROM + " DESC";

    private final Context mCtx;

    private static class DatabaseHelper extends SQLiteOpenHelper {

        DatabaseHelper(Context context) {
            super(context, DATABASE_NAME, null, DATABASE_VERSION);
            Logger.d(LOGTAG, "In constructor");
        }

        @Override
        public void onCreate(SQLiteDatabase db) {
            Logger.d(LOGTAG, "In onCreate");
            createInfoTable(db);
        
            String createMessageTableSQL = "create table " + MESSAGE_TABLE
                + " (" + 
                KEY_ID +        " integer primary key , " + 
                KEY_INFO_ID +   " integer   not null, " + 
                KEY_NUMBER +    " integer   not null, " + 
                KEY_TEXT +      " text      not null);";
        
            Logger.d(LOGTAG, "Executing createMessageTableSQL = " + createMessageTableSQL);
            db.execSQL(createMessageTableSQL);
        
            String createRestrictionTableSQL = "create table " + RESTRICTION_TABLE
                + " (" + 
                KEY_ID +          " integer primary key , " + 
                KEY_INFO_ID +     " integer   not null, " + 
                KEY_NUMBER +      " integer   not null, " + 
                KEY_TYPE +        " integer   not null, " + 
                KEY_TYPENAME +    " text      not null, " + 
                KEY_SPEED_LIMIT + " integer       null, " + 
                KEY_FROM +        " integer       null, " + 
                KEY_TO +          " integer       null);";
    
            Logger.d(LOGTAG, "Executing createRestrictionTableSQL = " + createRestrictionTableSQL);
            db.execSQL(createRestrictionTableSQL);
    
    
            String createRestrictionCommentTableSQL = "create table " + RESTRICTION_COMMENT_TABLE
                + " (" + 
                KEY_ID +                 " integer primary key , " +
                KEY_INFO_ID +            " integer   not null, " +
                KEY_RESTRICTION_NUMBER + " integer   not null, " + 
                KEY_NUMBER +             " integer   not null, " + 
                KEY_TEXT +               " text      not null);";
        
            Logger.d(LOGTAG, "Executing createRestrictionCommentTableSQL = " + createRestrictionCommentTableSQL);
            db.execSQL(createRestrictionCommentTableSQL);

            String createTimestampTableSQL = "create table " + TIMESTAMP_TABLE
                    + " (" + 
                    KEY_TYPE +      " integer    primary key, " + 
                    KEY_TIMESTAMP + " integer    not null );";

            Logger.d(LOGTAG, "Executing createTimestampTableSQL = " + createTimestampTableSQL);
            db.execSQL(createTimestampTableSQL);
            
            addPositionTable(db);
        }

        private void createInfoTable(SQLiteDatabase db) {
            String createInfoTableSQL = "create table " + INFORMATION_TABLE
                + " (" + 
                KEY_ID +        " integer primary key , " + 
                KEY_EXTERNALID +" text      not null, " + // TODO unique 
                KEY_HEADER +    " text      not null, " + 
                KEY_TYPE +      " integer   not null, " + 
                KEY_TYPENAME +  " text      not null, " + 
                KEY_SOURCE  +   " text          null, " + 
                KEY_FROM +      " integer       null, " + 
                KEY_TO +        " integer       null, " + 
                KEY_MODIFIED +  " integer       null, " + 
                KEY_LOCATION +  " text          null, " + 
                KEY_ROAD +      " text          null, " + 
                KEY_LONGITUDE + " real          null, " + 
                KEY_LATITUDE +  " real          null, " + 
                KEY_OBSOLETE +  " integer       null, " + 
                KEY_READ+       " integer       not null, " + 
                KEY_CATEGORY +  " integer       not null);";

            Logger.d(LOGTAG, "Executing createInfoTableSQL = " + createInfoTableSQL);
            db.execSQL(createInfoTableSQL);
        }

        @Override
        public void onUpgrade(SQLiteDatabase db, int oldVersion, int newVersion) {
            Logger.d(LOGTAG, "In onUpgrade oldVersion = " + oldVersion + " newVersion = " + newVersion);
            
            if (newVersion != DATABASE_VERSION)
                throw new RuntimeException("Database upgradescript must be reimplemented.");
            
            if (oldVersion == 2){
                upgrade2To3(db);
                return;
            } else if (oldVersion == 3){
                upgrade3To4(db);
                return;
            } else if (oldVersion == 4){
                upgrade4To5(db);
                return;
            }
            
            Logger.w(LOGTAG, "Upgrading database from version " + oldVersion + " to "
                    + newVersion + ", which will destroy all old data");
            
            db.execSQL("DROP TABLE IF EXISTS " + INFORMATION_TABLE);
            db.execSQL("DROP TABLE IF EXISTS " + TIMESTAMP_TABLE);
            db.execSQL("DROP TABLE IF EXISTS " + MESSAGE_TABLE);
            db.execSQL("DROP TABLE IF EXISTS " + RESTRICTION_TABLE);
            db.execSQL("DROP TABLE IF EXISTS " + RESTRICTION_COMMENT_TABLE);
            db.execSQL("DROP TABLE IF EXISTS " + POSITION_TABLE);
            
            Logger.d(LOGTAG, "In onUpgrade tables dropped");
            onCreate(db);
            Logger.d(LOGTAG, "In onUpgrade after calling onCreate");
        }

        private void upgrade4To5(SQLiteDatabase db) {
            db.execSQL("ALTER TABLE " + INFORMATION_TABLE + 
                    " ADD COLUMN " + KEY_CATEGORY +  " integer not null default 0");
            
            // Set categories for current records
            Cursor curs = db.query(INFORMATION_TABLE, // Table
                    new String[] { KEY_ID , KEY_TYPE }, // Columns
                    null, // Selection
                    null, // Selection arguments
                    null, // GROUP BY
                    null, // HAVING
                    null, // ORDER BY
                    null); // LIMIT

            while (curs.moveToNext()){
                long id   = curs.getLong(curs.getColumnIndexOrThrow(KEY_ID));
                int  type = curs.getInt(curs.getColumnIndexOrThrow(KEY_TYPE));

                ContentValues args = new ContentValues();
                args.put(KEY_CATEGORY, CategoryResolver.getCategoryByType(type));
                db.update(INFORMATION_TABLE, args, KEY_ID + " = " + id, null);
            }

            curs.close();
        }

        private void upgrade3To4(SQLiteDatabase db) {
            addPositionTable(db);
            
            recreateInfoTable(db);
        }

        private void recreateInfoTable(SQLiteDatabase db) {
            final String tempName = "tempTabel";
            db.execSQL("ALTER TABLE " + INFORMATION_TABLE + 
                    " RENAME TO " + tempName);
            
            createInfoTable(db);
            
            db.execSQL("INSERT INTO " + INFORMATION_TABLE + " SELECT * FROM " + tempName);
            
            db.execSQL("DROP TABLE " + tempName);
        }

        private void upgrade2To3(SQLiteDatabase db) {
            db.execSQL("ALTER TABLE " + INFORMATION_TABLE + 
                    " ADD COLUMN " + KEY_OBSOLETE +  " integer null");
            
            db.execSQL("ALTER TABLE " + INFORMATION_TABLE + 
                    " ADD COLUMN " + KEY_READ +  " integer not null default 0");
            
            
            ContentValues args = new ContentValues();
            args.put(KEY_TIMESTAMP, 4);

            args.put(KEY_TYPE, TIMESTAMPTYPE_LASTVERSION);

            if (db.insert(TIMESTAMP_TABLE, null, args) == -1){
                Logger.e(LOGTAG, "Could not insert last version");    
            } else {
                Logger.d(LOGTAG, "last version inserted");
            }
            
            
            // Mark all situations as read
            ContentValues args2 = new ContentValues();
            args2.put(KEY_READ, ReadState.Read.ordinal());
            args2.put(KEY_OBSOLETE, 0);
            db.update(INFORMATION_TABLE, args2, null, null);
        }
        
        private void addPositionTable(SQLiteDatabase db){
            String createPositionTableSQL = "create table " + POSITION_TABLE
            + " (" + 
            KEY_TYPE +      " integer    primary key, " + 
            KEY_LONGITUDE + " text       not null, " +
            KEY_LATITUDE +  " text       not null);";

            Logger.d(LOGTAG, "Executing createPositionTableSQL = " + createPositionTableSQL);
            db.execSQL(createPositionTableSQL);
        }
    }

    /**
     * Constructor - takes the context to allow the database to be
     * opened/created
     * 
     * @param ctx
     *            the Context within which to work
     */
    public InfoDbAdapter(Context ctx) {
        Logger.d(LOGTAG, "In constructor");
        this.mCtx = ctx;
    }

    /**
     * Open the notes database. If it cannot be opened, try to create a new
     * instance of the database. If it cannot be created, throw an exception to
     * signal the failure
     * 
     * @return this (self reference, allowing this to be chained in an
     *         initialization call)
     * @throws SQLException
     *             if the database could be neither opened or created
     */
    public InfoDbAdapter open() throws SQLException {
        mDbHelper = new DatabaseHelper(mCtx);
        mDb = mDbHelper.getWritableDatabase();
        return this;
    }

    public void close() {
        mDbHelper.close();
    }

    public void InsertOrUpdateInfo(TrafficInfo info) {
        Logger.d(LOGTAG, "In InsertOrUpdateInfo");
        
        if (info.getState() == ReadState.Unknown)
            Logger.w(LOGTAG, "ReadState is unknown");
        
        String externalId = info.getExternalId();
        ContentValues args = new ContentValues();
        args.put(KEY_HEADER, info.getHeader());
        args.put(KEY_TYPENAME, info.getTypeName());
        args.put(KEY_LOCATION, info.getLocation());
        args.put(KEY_TYPE, info.getType());
        args.put(KEY_CATEGORY, info.getCategory());
        args.put(KEY_SOURCE, info.getSource());
        args.put(KEY_ROAD, info.getRoad());
        args.put(KEY_LONGITUDE, info.getLongitude());
        args.put(KEY_LATITUDE, info.getLatitude());
        args.put(KEY_FROM, dateToLong(info.getFrom()));
        args.put(KEY_TO, dateToLong(info.getTo()));
        args.put(KEY_MODIFIED, dateToLong(info.getModified()));
        args.put(KEY_OBSOLETE, 0);
        args.put(KEY_READ, info.getState().ordinal());

        long infoId = 0; // TODO constant unknown
        
        if (mDb.update(INFORMATION_TABLE, args, KEY_EXTERNALID + " = '" + externalId + "'",
                null) > 0){
            Logger.d(LOGTAG, "In InsertOrUpdateInfo data updated");    
        } else {
            args.put(KEY_EXTERNALID, externalId);
            
            infoId = mDb.insert(INFORMATION_TABLE, null, args); 
            if (infoId == -1){
                Logger.e(LOGTAG, "Could not insert info");
            } else {
                Logger.d(LOGTAG, "In InsertOrUpdateInfo data inserted");
            }
        }
        
        if (infoId == 0) // TODO constant unknown
            infoId = getIdByExternalId(info.getExternalId());
        
        List<String> comments = info.getComments();
        int numberOfComments = comments.size();
        
        Logger.d(LOGTAG, "In InsertOrUpdateInfo inserting " + numberOfComments + " comments");
        
        for (int i = 0 ; i < numberOfComments ; i++){
            String comment = comments.get(i);
            insertOrUpdateComment(infoId, i, comment);
        }
        
        deleteObsoleteComments(infoId, numberOfComments - 1);
       
        List<Restriction> restrictions = info.getRestricions();
        int numberOfRestrictions = restrictions.size();
        
        Logger.d(LOGTAG, "In InsertOrUpdateInfo inserting " + numberOfRestrictions + " restrictions");
        
        for (int i = 0; i < numberOfRestrictions ; i++){
            Restriction restriction = restrictions.get(i);
            insertOrUpdateRestriction(infoId, i, restriction);

            List<String> restrictionComments = restriction.getComments();
            int numberOfRestrictionComments = restrictionComments.size();
            
            Logger.d(LOGTAG, "In InsertOrUpdateInfo inserting " + numberOfRestrictionComments + " restrictioncomments");
            
            for (int j = 0; j < numberOfRestrictionComments ; j++){
                String restrictionComment = restrictionComments.get(j);
                insertOrUpdateRestrictionComment(infoId, i, j, restrictionComment);
            }
            
            deleteObsoleteRestrictonComments(infoId, i, numberOfRestrictionComments -1);
        }
        
        deleteObsoleteRestrictons(infoId, numberOfRestrictions - 1);
    }

    public long getIdByExternalId(String externalId) {
        Logger.d(LOGTAG, "in getIdByExternalId");
        long ret = 0; // TODO use  constant unknown ...
        
        Cursor curs = mDb.query(INFORMATION_TABLE, // Table
                new String[] { KEY_ID }, // Columns
                KEY_EXTERNALID + " = '" + externalId + "'", // Selection
                null, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                null, // ORDER BY
                "1"); // LIMIT

        if (curs.moveToNext())
            ret = curs.getLong(curs.getColumnIndexOrThrow(KEY_ID));

        curs.close();

        return ret;

    }

    private void deleteObsoleteRestrictonComments(long infoId, int restrivtionNr, int numberOfComment) {
        Logger.d(LOGTAG, "in deleteObsoleteRestrictonComments");

        String filter = 
            KEY_INFO_ID +            " = " + infoId + " AND " + 
            KEY_RESTRICTION_NUMBER + " = " + infoId + " AND " + 
            KEY_NUMBER +             " > " + numberOfComment;

        mDb.delete(RESTRICTION_COMMENT_TABLE, filter, null);
        Logger.d(LOGTAG, "in deleteObsoleteRestrictonComments after delete");
    }

    private void insertOrUpdateRestrictionComment(long infoId, int restrictionNr, int commentNr,
            String comment) {
        Logger.d(LOGTAG, "In insertOrUpdateRestrictionComment");
        ContentValues args = new ContentValues();

        args.put(KEY_TEXT, comment);

        String whereClause = 
            KEY_INFO_ID            + " = " + infoId + " AND " + 
            KEY_RESTRICTION_NUMBER + " = " + restrictionNr + " AND " + 
            KEY_NUMBER             + " = " + commentNr;
        
        if (mDb.update(RESTRICTION_COMMENT_TABLE, args, whereClause ,
                null) > 0){
            Logger.d(LOGTAG, "In insertOrUpdateRestrictionComment data updated");    
        } else {
            args.put(KEY_INFO_ID, infoId);
            args.put(KEY_RESTRICTION_NUMBER, restrictionNr);
            args.put(KEY_NUMBER, commentNr);
    
            if (mDb.insert(RESTRICTION_COMMENT_TABLE, null, args) == -1){
                Logger.e(LOGTAG, "Could not insert comment");
            } else {
                Logger.d(LOGTAG, "In insertOrUpdateRestrictionComment data inserted");
            }
        }
    }

    private void deleteObsoleteRestrictons(long infoId, int maxNumber) {
        deleteObsoleteObjects(infoId, maxNumber, RESTRICTION_TABLE);
    }

    private void deleteObsoleteComments(long infoId, int maxNumber) {
        deleteObsoleteObjects(infoId, maxNumber, MESSAGE_TABLE);    
    }
    
    private void deleteObsoleteObjects(long infoId, int maxNumber, String tableName) {
        Logger.d(LOGTAG, "in deleteObsoleteObjects");

        String filter = KEY_INFO_ID + " = " + infoId + " AND " + KEY_NUMBER + " > " + maxNumber;

        mDb.delete(tableName, filter, null);
        Logger.d(LOGTAG, "in deleteObsoleteObjects after delete");
    }

    private void insertOrUpdateRestriction(long infoId, int nr, Restriction restriction) {
        Logger.d(LOGTAG, "In insertOrUpdateRestriction");
        ContentValues args = new ContentValues();

        args.put(KEY_TYPE, restriction.getType());
        args.put(KEY_TYPENAME, restriction.getTypeName());
        args.put(KEY_SPEED_LIMIT, restriction.getSpeedLimit());
        args.put(KEY_FROM, dateToLong(restriction.getFrom()));
        args.put(KEY_TO, dateToLong(restriction.getTo()));
        
        String whereClause = 
            KEY_INFO_ID + " = " + infoId + " AND " + 
            KEY_NUMBER  + " = " + nr;
        
        if (mDb.update(RESTRICTION_TABLE, args, whereClause ,
                null) > 0){
            Logger.d(LOGTAG, "In insertOrUpdateRestriction data updated");    
        } else {
            args.put(KEY_INFO_ID, infoId);
            args.put(KEY_NUMBER, nr);
    
            if (mDb.insert(RESTRICTION_TABLE, null, args) == -1){
                Logger.e(LOGTAG, "Could not insert comment");
            } else {
                Logger.d(LOGTAG, "In insertOrUpdateRestriction data inserted");
            }
        }
    }
    
    private void insertOrUpdateComment(long infoId, int nr, String comment) {
        Logger.d(LOGTAG, "In insertOrUpdateComment");
        ContentValues args = new ContentValues();

        args.put(KEY_TEXT, comment);

        String whereClause = 
            KEY_INFO_ID + " = " + infoId + " AND " + 
            KEY_NUMBER  + " = " + nr;
        
        if (mDb.update(MESSAGE_TABLE, args, whereClause ,
                null) > 0){
            Logger.d(LOGTAG, "In insertOrUpdateComment data updated");    
        } else {
            args.put(KEY_INFO_ID, infoId);
            args.put(KEY_NUMBER, nr);
    
            if (mDb.insert(MESSAGE_TABLE, null, args) == -1){
                Logger.e(LOGTAG, "Could not insert comment");
            } else {
                Logger.d(LOGTAG, "In insertOrUpdateComment data inserted");
            }
        }
    }

    private Long dateToLong(Date date) {
        return date == null ? null : date
                .getTime();
    }

    public Cursor fetchAllActiveSituations() {
        Logger.d(LOGTAG, "In fetchAllActiveSituations");
        String filter = getActiveSituationsFilter();
        return getTrafficInfoCursorByFilter(filter, ORDERBY, null);
    }

    public Cursor fetchAllActiveSituationsWithPosition() {
        Logger.d(LOGTAG, "In fetchAllActiveSituations");
        String filter = getActiveSituationsFilter();
        if (filter.length() > 0)
            filter += " AND ";
        filter += "(" + KEY_LATITUDE + " IS NOT NULL OR " + KEY_LATITUDE + " <> 0) AND ";
        filter += "(" + KEY_LONGITUDE + " IS NOT NULL OR " + KEY_LONGITUDE + " <> 0)";
        return getTrafficInfoCursorByFilter(filter, ORDERBY, null);
    }

    public Cursor fetchAllFutureSituations() {
        Logger.d(LOGTAG, "In fetchAllFutureSituations");
        String filter = getFutureSituationsFilter();
        return getTrafficInfoCursorByFilter(filter, REVERSE_ORDERBY, null);
    }

    private Cursor getTrafficInfoCursorByFilter(String filter, String orderBy, String limit) {
        Logger.d(LOGTAG, "In getTrafficInfoCursorByFilter filter = " + filter);
        return mDb.query(INFORMATION_TABLE, // Table
                new String[] { 
                    KEY_ID,
                    KEY_EXTERNALID,
                    KEY_FROM, 
                    KEY_TO ,
                    KEY_HEADER,
                    KEY_TYPE,
                    KEY_TYPENAME,
                    KEY_SOURCE,
                    KEY_MODIFIED,
                    KEY_LOCATION,
                    KEY_ROAD,
                    KEY_LONGITUDE,
                    KEY_LATITUDE,
                    KEY_READ,
                    KEY_CATEGORY}, // Columns
                filter, // Selection
                null, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                orderBy, // ORDER BY
                limit); // LIMIT
    }

    private TrafficInfo extractInfoFromCursor(Cursor curs) {
        Logger.d(LOGTAG, "In extractInfoFromCursor");
        String externalId = curs.getString(curs.getColumnIndexOrThrow(KEY_EXTERNALID));
        Date from = extractDate(curs, KEY_FROM);
        Date to = extractDate(curs, KEY_TO); 
        String header = curs.getString(curs.getColumnIndexOrThrow(KEY_HEADER)); 
        int type = curs.getInt(curs.getColumnIndexOrThrow(KEY_TYPE));
        String typeName = curs.getString(curs.getColumnIndexOrThrow(KEY_TYPENAME));
        String source = curs.getString(curs.getColumnIndexOrThrow(KEY_SOURCE));
        String location = curs.getString(curs.getColumnIndexOrThrow(KEY_LOCATION));
        String road = curs.getString(curs.getColumnIndexOrThrow(KEY_ROAD));
        Date modified = extractDate(curs, KEY_MODIFIED);;
        Double longitude = curs.getDouble(curs.getColumnIndexOrThrow(KEY_LONGITUDE));
        Double latitude = curs.getDouble(curs.getColumnIndexOrThrow(KEY_LATITUDE));
        int category = curs.getInt(curs.getColumnIndexOrThrow(KEY_CATEGORY));
        int stateInt = curs.getInt(curs.getColumnIndexOrThrow(KEY_READ));
        ReadState state = ReadState.values()[ stateInt];
        TrafficInfo info = new TrafficInfo(
                externalId,
                header, 
                type, 
                category,
                typeName,
                source, 
                location,
                road,
                modified,
                longitude,
                latitude,
                from, 
                to, 
                state,
                new ArrayList<String>(),
                new ArrayList<Restriction>());
        
        Logger.d(LOGTAG, "returning from extractInfoFromCursor");
        return info;
    }

    private Date extractDate(Cursor curs, String field) {
        Logger.d(LOGTAG, "in extractDate");
        Date date;
        int columnIndex = curs.getColumnIndexOrThrow(field);
        if (curs.isNull(columnIndex))
            date = null;
        else
            date = new Date(curs.getLong(columnIndex));
        return date;
    }

    public void markInformatiomObsoleteNotInList(List<Long> test) {
        Logger.d(LOGTAG, "in markInformatiomObsoleteNotInList");
        String filter = "";
        
        if (test.size() != 0) {
            filter += KEY_ID + " NOT IN (";
    
            for (Long id : test)
                filter += id + ", ";
    
            filter = filter.substring(0, filter.length() - 2) + ")";
        }
        
        markObsoleteByFilter(filter);
        Logger.d(LOGTAG, "in deleteInformationNotInList after delete");
    }

    private void markObsoleteByFilter(String filter) {
        Logger.d(LOGTAG, "in markObsoleteByFilter");
        Logger.d(LOGTAG, "filter = " + filter);
        
        if (filter.length() > 0)
            filter += " AND ";
        
        filter += KEY_OBSOLETE + " = 0";
        
        ContentValues values = new ContentValues();
        long obsoleteTimestamp = (new Date()).getTime();
        values.put(KEY_OBSOLETE, obsoleteTimestamp);
        
        int numberUpdated = mDb.update(INFORMATION_TABLE, values, filter, null);
        Logger.d(LOGTAG, numberUpdated + " situations marked as obsolete.");
    }
    
    public void markInformatiomObsoleteOutsideArea(
            Double longMin, 
            Double longMax, 
            Double latMin, 
            Double latMax ) {
        Logger.d(LOGTAG, "in markInformatiomObsoleteOutsideArea");

        String filter = "(" + 
        KEY_LONGITUDE + " < " + longMin + " OR " + 
        KEY_LONGITUDE + " > " + longMax + " OR " + 
        KEY_LATITUDE  + " < " + latMin  + " OR " + 
        KEY_LATITUDE  + " > " + latMax  + ")";

        markObsoleteByFilter(filter);
        Logger.d(LOGTAG, "in markInformatiomObsoleteOutsideArea after delete");
    }


    public void setUpdateTimestamp(Date timestamp) {
        Logger.d(LOGTAG, "in setUpdateTimestamp");
        updateTimestampByType(timestamp, TIMESTAMPTYPE_LASTUPDATE);
    }

    public void setServerTimestamp(Date timestamp) {
        Logger.d(LOGTAG, "in setServerTimestamp");
        updateTimestampByType(timestamp, TIMESTAMPTYPE_SERVERTIME);
    }
    
    public Date getServerTimestamp() {
        Logger.d(LOGTAG, "in getServerTimestamp");
        return getTimestampByType(TIMESTAMPTYPE_SERVERTIME);
    }
    
    public void setCleanupTimestamp(Date timestamp) {
        Logger.d(LOGTAG, "in setCleanupTimestamp");
        updateTimestampByType(timestamp, TIMESTAMPTYPE_LASTCLEANUPTIME);
    }
    
    public Date getClenupTimestamp() {
        Logger.d(LOGTAG, "in getClenupTimestamp");
        return getTimestampByType(TIMESTAMPTYPE_LASTCLEANUPTIME);
    }

    public long getLastUpdateRegion() {
        Logger.d(LOGTAG, "in getLastUpdateRegion");
        return getLong(TIMESTAMPTYPE_LASTUPATECOUNTY, -1L);
    }

    public void setLastUpdateRegion(long countyId) {
        Logger.d(LOGTAG, "in setLastUpdateRegion");
        setLongValue(TIMESTAMPTYPE_LASTUPATECOUNTY,  countyId);
    }
    
    
    public long getLastVersion() {
        Logger.d(LOGTAG, "in getLastVersion");
        return getLong(TIMESTAMPTYPE_LASTVERSION, 0L);
    }

    public void setLastVersion(long version) {
        Logger.d(LOGTAG, "in setLastVersion");
        setLongValue(TIMESTAMPTYPE_LASTVERSION,  version);
    }

    private void updateTimestampByType(Date timestamp, int type) {
        Logger.d(LOGTAG, "in updateTimestampByType");
        setLongValue(type,  timestamp.getTime());
    }

    private void setLongValue(int type, long value) {
        ContentValues args = new ContentValues();
        args.put(KEY_TIMESTAMP, value);
        String filter = KEY_TYPE + " = " + type;

        if (mDb.update(TIMESTAMP_TABLE, args, filter, null) > 0){
            Logger.d(LOGTAG, "Timestamp updated");
            return;
        }
        
        args.put(KEY_TYPE, type);

        if (mDb.insert(TIMESTAMP_TABLE, null, args) == -1){
            Logger.e(LOGTAG, "Could not insert info");    
        } else {
            Logger.d(LOGTAG, "Timestamp inserted");
        }
    }

    public Date getUpdateTimestamp() {
        Logger.d(LOGTAG, "in getUpdateTimestamp");
        return getTimestampByType(TIMESTAMPTYPE_LASTUPDATE);
    }

    private Date getTimestampByType(int type) {
        Logger.d(LOGTAG, "in getTimestampByType");
        Long time = getLong(type, 0L);

        if (time == 0L)
            return null;
        
        return new Date(time);
    }

    private Long getLong(int type, long defaultValue) {
        String filter = KEY_TYPE + " = " + type; 
        Cursor curs = mDb.query(TIMESTAMP_TABLE, // Table
                new String[] { KEY_TIMESTAMP }, // Columns
                filter, // Selection
                null, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                null); // ORDER BY
        Long time = defaultValue;
        if (curs.moveToNext())
            time = curs.getLong(curs
                    .getColumnIndexOrThrow(KEY_TIMESTAMP));

        curs.close();
        return time;
    }
    
    public Position getPoint(int type, Position defaultValue) {
        String filter = KEY_TYPE + " = " + type; 
        Cursor curs = mDb.query(POSITION_TABLE, // Table
                new String[] { KEY_LONGITUDE, KEY_LATITUDE }, // Columns
                filter, // Selection
                null, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                null); // ORDER BY
        Position point = defaultValue;
        if (curs.moveToNext()){
            point.Longitude = curs.getDouble(curs
                    .getColumnIndexOrThrow(KEY_LONGITUDE));
            point.Latitude = curs.getDouble(curs
                    .getColumnIndexOrThrow(KEY_LATITUDE));
        }
        curs.close();
        return point;
    }
    
    public void setPoint(int type, Position value) {
        ContentValues args = new ContentValues();
        args.put(KEY_LONGITUDE, value.Longitude);
        args.put(KEY_LATITUDE, value.Latitude);
        String filter = KEY_TYPE + " = " + type;

        if (mDb.update(POSITION_TABLE, args, filter, null) > 0){
            Logger.d(LOGTAG, "Position updated");
            return;
        }
        
        args.put(KEY_TYPE, type);

        if (mDb.insert(POSITION_TABLE, null, args) == -1){
            Logger.e(LOGTAG, "Could not insert position");    
        } else {
            Logger.d(LOGTAG, "Position inserted");
        }
    }

    public TrafficInfo getInformationById(long id) {
        Logger.d(LOGTAG, "in getInformationById");
        Cursor cursor = getTrafficInfoCursorByFilter(KEY_ID + " = " + id, null, "1");
        TrafficInfo ret = null;

        if (cursor.moveToNext())
            ret = extractInfoFromCursor(cursor);

        cursor.close();
        
        if (ret == null)
            return null;
        
        List<String> commentList = ret.getComments();
        Cursor commentsCursor = getCommentCursorByInfoId(id);

        while (commentsCursor.moveToNext()){
            String comment = commentsCursor.getString(
                    commentsCursor.getColumnIndexOrThrow(KEY_TEXT));
            
            commentList.add(comment);
        }
        
        commentsCursor.close();

        List<Restriction> restrictionList = ret.getRestricions();
        Cursor restrictionCursor = getRestrictionCursorByInfoId(id);

        while (restrictionCursor.moveToNext()){
            Restriction restriction = extractRestrictionFromCursor(restrictionCursor);
            Cursor restrictionCommentCursor = getRestrictionCommentCursor(id, restriction.getNumber());
            
            List<String> restrictionCommentList = restriction.getComments();
            while (restrictionCommentCursor.moveToNext()){
                String comment = restrictionCommentCursor.getString(
                        restrictionCommentCursor.getColumnIndexOrThrow(KEY_TEXT));
                
                restrictionCommentList.add(comment);
            }
            
            restrictionCommentCursor.close();
            
            restrictionList.add(restriction);
        }
        
        restrictionCursor.close();

        return ret;
    }
    
    private Cursor getRestrictionCommentCursor(long infoId, int number) {
        Logger.d(LOGTAG, "In getRestrictionCommentCursor id = " + infoId);
        return mDb.query(RESTRICTION_COMMENT_TABLE, // Table
                new String[] {KEY_TEXT}, // Columns
                KEY_INFO_ID + " = " + infoId + " AND " + 
                KEY_RESTRICTION_NUMBER + " = " + number, // Selection
                null, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                KEY_NUMBER); // ORDER BY
    }

    private Restriction extractRestrictionFromCursor(Cursor curs) {
        Logger.d(LOGTAG, "In extractRestrictionFromCursor");
        Date from = extractDate(curs, KEY_FROM);
        Date to = extractDate(curs, KEY_TO);  
        int type = curs.getInt(curs.getColumnIndexOrThrow(KEY_TYPE));
        String typeName = curs.getString(curs.getColumnIndexOrThrow(KEY_TYPENAME));
        int speedLimit = curs.getInt(curs.getColumnIndexOrThrow(KEY_SPEED_LIMIT));
        int number = curs.getInt(curs.getColumnIndexOrThrow(KEY_NUMBER));
        
        Restriction restriction = new Restriction(
                from, 
                to,
                number,
                type, 
                typeName,
                speedLimit, 
                new ArrayList<String>());
        
        Logger.d(LOGTAG, "returning from extractRestrictionFromCursor");
        return restriction;
    }

    private Cursor getRestrictionCursorByInfoId(long id) {
        Logger.d(LOGTAG, "In getRestrictionCursorByInfoId id = " + id);
        return mDb.query(RESTRICTION_TABLE, // Table
                new String[] {
                KEY_TYPE,
                KEY_TYPENAME,
                KEY_SPEED_LIMIT,
                KEY_FROM,
                KEY_TO,
                KEY_INFO_ID, 
                KEY_NUMBER
                }, // Columns
                KEY_INFO_ID + " = " + id, // Selection
                null, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                KEY_NUMBER); // ORDER BY
    }

    private Cursor getCommentCursorByInfoId(long id) {
        Logger.d(LOGTAG, "In getCommentCursorByInfoId id = " + id);
        return mDb.query(MESSAGE_TABLE, // Table
                new String[] {KEY_TEXT}, // Columns
                KEY_INFO_ID + " = " + id, // Selection
                null, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                KEY_NUMBER); // ORDER BY
    }
    
    private String getCommonSituationsFilter(){
        String filter = "";
        // TODO Or KEY_OBSOLETE is NULL
        filter += KEY_OBSOLETE + " = 0";
        filter += " AND " + KEY_READ + " < " + ReadState.UserDeleted.ordinal();
        PreferenceHandler handler = new PreferenceHandler();
        long categoryMask = handler.getCategoryMask(mCtx);
        filter += " AND " + KEY_CATEGORY + " & " + categoryMask + " <> 0";
        return filter; 
    }
    
    private String getActiveSituationsFilter(){
        long now = new Date().getTime();
        String filter = "(" + KEY_FROM + " < " + now + " OR " + KEY_FROM + " IS NULL) AND " +
                        "(" + KEY_TO   + " > " + now + " OR " + KEY_TO   + " IS NULL)";
        filter += " AND " + getCommonSituationsFilter();
        return filter;
    }
    
    private String getFutureSituationsFilter(){
        long now = new Date().getTime();
        String filter = KEY_FROM + " > " + now;
        filter += " AND " + getCommonSituationsFilter();
        return filter; 
    }

    public TrafficInfo getNewest(){
        Logger.d(LOGTAG, "in getNewest");
        TrafficInfo ret = null;
        
        String filter = getActiveSituationsFilter();
        
        Cursor cursor = getTrafficInfoCursorByFilter(filter, ORDERBY, "1");
        

        if (cursor.moveToNext())
        {
            ret = extractInfoFromCursor(cursor);
        }

        cursor.close();

        return ret;
    }
    
    public int getNumberOfMessages(){
        Logger.d(LOGTAG, "in getNumberOfMessages");
        String filter = getActiveSituationsFilter();
        return countSituationsByFilter(filter);
    }

    public int getNumberOfUnreadSituations(){
        Logger.d(LOGTAG, "in getNumberOfUnreadSituations");
        int newConst = ReadState.New.ordinal();
        int modifiedConst = ReadState.Modified.ordinal();
        String filter = KEY_READ + " IN (" + newConst + ", " + modifiedConst + ")";
        filter += " AND ";
        filter += getActiveSituationsFilter();
        return countSituationsByFilter(filter);
    }

    private int countSituationsByFilter(String filter) {
        Logger.d(LOGTAG, "in countSituationsByFilter");
        int ret = 0;
        Cursor curs = mDb.query(INFORMATION_TABLE, // Table
                new String[] { "Count(*)" }, // Columns
                filter, // Selection
                null, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                null); // ORDER BY

        if (curs.moveToNext())
            ret = curs.getInt(0);

        curs.close();

        return ret;
    }
    
    public class SituationStatus{
        private final ReadState mReadState;
        private final Date mTimestamp;
        
        public SituationStatus(ReadState readState, Date timeStamp){
            mReadState = readState;
            mTimestamp = timeStamp;
        }
        
        public ReadState getReadState(){
            return mReadState;
        }
        
        public Date getTimeStamp(){
            return mTimestamp;
        }
    }
    
    public SituationStatus findCurrentStatus(String externalId){
        Logger.d(LOGTAG, "in findCurrentStatus");

        String[] selectionArgs = new String[] { externalId }; 
        
        Cursor curs = mDb.query(INFORMATION_TABLE, // Table
                new String[] { KEY_MODIFIED, KEY_READ }, // Columns
                KEY_EXTERNALID + " = ?", // Selection
                selectionArgs, // Selection arguments
                null, // GROUP BY
                null, // HAVING
                null, // ORDER BY
                null); // LIMIT
        
        SituationStatus returnValue;

        if (curs.moveToNext())
        {
            int lastReadStateInt = curs.getInt(curs.getColumnIndexOrThrow(KEY_READ));
            ReadState lastReadState = ReadState.values()[lastReadStateInt];
            Date timestamp = extractDate(curs, KEY_MODIFIED);
            returnValue = new SituationStatus(lastReadState, timestamp);
            
        } else {
            returnValue = null;
        }
        curs.close();
        return returnValue;
    }
    
    public void markSiuationsAsRead(List<Long> ids, Context context){
        Logger.d(LOGTAG, "in markSiuationsAsRead");
        setReadStateForSituations(ids, context, ReadState.Read);
    }

    public void markSiuationDeleted(Long id, Context context){
        Logger.d(LOGTAG, "in markSiuationDeleted");
        List<Long> list = new ArrayList<Long>(1);
        list.add(id);
        setReadStateForSituations(list, context, ReadState.UserDeleted);
    }
    
    public void markSiuationsDeleted(List<Long> ids, Context context){
        Logger.d(LOGTAG, "in markSiuationsDeleted");
        setReadStateForSituations(ids, context, ReadState.UserDeleted);
    }

    private void setReadStateForSituations(List<Long> ids, Context context,
            ReadState state) {
        Logger.d(LOGTAG,"in setReadStateForSituations");
        int numberOfSituations = ids.size();
        Logger.d(LOGTAG,"in setReadStateForSituations " + numberOfSituations + " ids in list");
        
        String idsString = "";
        for (int i = 0; i < numberOfSituations ; i++ ){
            long id = ids.get(i);
            if (i > 0)
                idsString += ", ";
            idsString += id;
        }
        ContentValues args = new ContentValues();
        args.put(KEY_READ, state.ordinal());
        int numberUpdated = mDb.update(INFORMATION_TABLE, args, KEY_ID + " IN (" + idsString + ")", null);
        Logger.d(LOGTAG, "updated " + numberUpdated + " situations");
        
        // Tell the widget to update number of unread
        InfoFetchService.sendUpdateIntent(null, context);
        
        // Update the notification
        InfoFetchService.updateUserNotification(context, this, false);
    }
    
    public void deleteOldObsoleteSituations(Date olderThan){
        Logger.d(LOGTAG, "in deleteOldObsoleteSituations");
        Logger.d(LOGTAG, "in olderThan = " + olderThan.toString());
        String filter = KEY_OBSOLETE + " > 0 AND " + KEY_OBSOLETE + " < " + olderThan.getTime();
        int numberDeleted = mDb.delete(INFORMATION_TABLE, filter, null);
        Logger.d(LOGTAG, "deleted " + numberDeleted + " situations");
        
        deleteOrphans(MESSAGE_TABLE);
        deleteOrphans(RESTRICTION_TABLE);
        deleteOrphans(RESTRICTION_COMMENT_TABLE);
    }

    private void deleteOrphans(String tableName) {
        Logger.d(LOGTAG, "in deleteOrphans");
        String filter = "NOT EXISTS (SELECT 1 FROM " + INFORMATION_TABLE + " i WHERE i." + KEY_ID + " = " + tableName + "." + KEY_INFO_ID + ")";
        int numberDeleted = mDb.delete(tableName, filter, null);
        Logger.d(LOGTAG, numberDeleted + " rows deleted from table " + tableName);
    }

    public Cursor fetchSituation(String externalId) {
        Logger.d(LOGTAG, "in fetchSituation");
        String filter = KEY_EXTERNALID + " = '" + externalId + "'";
        return getTrafficInfoCursorByFilter(filter, null, null);
    }
}
