/*
    This source file is part of the trafikinfo for android project
    Copyright (C) 2009-2010  Peter Björkman


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package se.peterbjorkman.android.trafikinfo;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoException;
import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoException.ErrorCodes;
import android.app.Notification;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Service;
import android.app.PendingIntent.CanceledException;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.database.SQLException;
import android.location.Location;
import android.os.Binder;
import android.os.Bundle;
import android.os.Handler;
import android.os.HandlerThread;
import android.os.IBinder;
import android.os.Looper;
import android.os.Message;
import android.os.PowerManager;
import android.preference.PreferenceManager;
import android.provider.Settings.Secure;
import android.util.FloatMath;
import android.util.Log;

public class InfoFetchService extends Service {
    private static final String LOGTAG = InfoFetchService.class.getName();

    private volatile Looper mServiceLooper;
    private volatile ServiceHandler mServiceHandler;

    private InfoDbAdapter mDbAdapter;

    private Date mLastStartTime;

    private static final int MESSAGE_COMPLETE = 1;

    private static final String MESSAGE_TIMESTAMP = "timestamp";

    public static final String UPDATE_INTENT = "se.peterbjorkman.android.trafikinfo.dataupdate";
    public static final String UPDATE_INTENT_EXTRA_STATUS = "status";
    public static final String UPDATE_INTENT_EXTRA_DONE = "done";
    public static final String UPDATE_INTENT_EXTRA_TOTAL = "total";
    public static final String UPDATE_INTENT_EXTRA_EXCEPTION = "exception";
    
    private static final int NOTIFICATION_ID = 1;

    final List<IDataObserver> mCallbacks = new ArrayList<IDataObserver>();

    public static final String LOCK_NAME_STATIC = "se.peterbjorkman.andriod.trafikinfo.fetchinfopservice.powerlock";
    
    private static PowerManager.WakeLock lockStatic = null;

    public static void acquireStaticLock(Context context) {
        Log.d(LOGTAG, "in acquireStaticLock");
        // TODO Use some timeout for the lock??
        getLock(context).acquire();
    }

    synchronized private static PowerManager.WakeLock getLock(Context context) {
        Log.d(LOGTAG, "in getLock");
        if (lockStatic == null) {
            Log.d(LOGTAG, "in creating new wake lock");
            PowerManager mgr = (PowerManager) context
                    .getSystemService(Context.POWER_SERVICE);

            lockStatic = mgr.newWakeLock(
                    PowerManager.PARTIAL_WAKE_LOCK,
                    LOCK_NAME_STATIC);
            
            lockStatic.setReferenceCounted(false);
        }

        return (lockStatic);
    }

    /**
     * Class for clients to access. Because we know this service always runs in
     * the same process as its clients, we don't need to deal with IPC.
     */
    public class LocalBinder extends Binder {
        InfoFetchService getService() {
            return InfoFetchService.this;
        }
    }

    @Override
    public void onCreate() {
        Log.d(LOGTAG, "In onCreate");
        HandlerThread thread = new HandlerThread("ServiceStartArguments");
        thread.start();

        mServiceLooper = thread.getLooper();
        mServiceHandler = new ServiceHandler(mServiceLooper);
        mDbAdapter = new InfoDbAdapter(this);
    }

    @Override
    public void onDestroy() {
        Log.d(LOGTAG, "In onDestroy");
        super.onDestroy();
    }

    @Override
    public void onStart(Intent intent, int startId) {
        Log.d(LOGTAG, "In onStart");
        super.onStart(intent, startId);
        mLastStartTime = new Date();
        startCompleteUpdate();
    }

    @Override
    public IBinder onBind(Intent intent) {
        Log.d(LOGTAG, "In onBind");
        return mBinder;
    }

    // This is the object that receives interactions from clients. See
    // RemoteService for a more complete example.
    private final IBinder mBinder = new LocalBinder();

    public void startCompleteUpdate() {
        Log.d(LOGTAG, "Complete update requested");
        Message msg = mServiceHandler.obtainMessage();
        msg.what = MESSAGE_COMPLETE;
        msg.getData().putSerializable(MESSAGE_TIMESTAMP, new Date());
        mServiceHandler.sendMessage(msg);
    }

    private final class ServiceHandler extends Handler {
        public ServiceHandler(Looper looper) {
            super(looper);
        }

        private Date lastCompleteUpdate = null;

        @Override
        public void handleMessage(Message msg) {
            Log.d(LOGTAG, "in handleMessage");

            // TODO check this cast somehow??
            Date messageTimeStamp = (Date) msg.getData().getSerializable(
                    MESSAGE_TIMESTAMP);

            if (lastCompleteUpdate != null
                    && messageTimeStamp.compareTo(lastCompleteUpdate) < 0) {
                Log.i(LOGTAG, "Ignore uppdate. Complete update already done.");
                return;
            }

            if (msg.what == MESSAGE_COMPLETE)
                lastCompleteUpdate = new Date();
            else {
                Log.w(LOGTAG, "Unknown message type: " + msg.what);
                return;
            }

            try {
                mDbAdapter.open();

                PreferenceHandler prefHandler = new PreferenceHandler();
                
                int regionId = getRegionIdFromPreference();

                Date lastUpdateTimestamp = null;
                Boolean regionChanged = regionId != mDbAdapter.getLastUpdateRegion(); 
                if (!regionChanged)
                    lastUpdateTimestamp = mDbAdapter.getServerTimestamp();
                else
                    mDbAdapter.setLastUpdateRegion(regionId);
                
                String deviceId = Secure.getString(getContentResolver(), Secure.ANDROID_ID);
               
                if (deviceId == null || deviceId.contentEquals(""))
                    deviceId = "null";
                int versionCode = WelcomeDialogBuilder.getCurrentVersionCode(getBaseContext());
                
                IDataFetcher fetcher;
                if (regionId == 0){
                    final Double radius = prefHandler.getRadius(InfoFetchService.this) * 10000.0;
                    int minAccurency = (int)(radius / 10.0);
                    Position position = getCurrentPosition(minAccurency);

                    // TODO: Add Accuracy to the preferred radius
                    final Double equatorLength = 40075035.0;
                    
                    Double deltaLat = 360 * radius/equatorLength; 
                    Float cosLat = FloatMath.cos(new Float(position.Latitude * 3.14159 / 180.0));
                    Double deltaLong = 360 * radius/(equatorLength * cosLat);

                    Position minFetch = new Position(
                            position.Longitude - deltaLong, 
                            position.Latitude - deltaLat);

                    Position maxFetch = new Position(
                            position.Longitude + deltaLong, 
                            position.Latitude + deltaLat);
    
                    Position lastFetchMin = mDbAdapter.getPoint(InfoDbAdapter.POSITIONTYPE_LASTFETCHMIN, new Position());
                    Position lastFetchMax = mDbAdapter.getPoint(InfoDbAdapter.POSITIONTYPE_LASTFETCHMAX, new Position());
                    
                    mDbAdapter.setPoint(InfoDbAdapter.POSITIONTYPE_LASTFETCHMIN, minFetch);
                    mDbAdapter.setPoint(InfoDbAdapter.POSITIONTYPE_LASTFETCHMAX, maxFetch);
                    
                    mDbAdapter.markInformatiomObsoleteOutsideArea(minFetch.Longitude, 
                            maxFetch.Longitude, 
                            minFetch.Latitude, 
                            maxFetch.Latitude);
                    
                    fetcher = new TrafikInfoNuCoordFetcher(
                            deviceId,
                            versionCode,
                            minFetch.Longitude, 
                            maxFetch.Longitude, 
                            minFetch.Latitude, 
                            maxFetch.Latitude, 
                            lastFetchMin.Longitude, 
                            lastFetchMax.Longitude, 
                            lastFetchMin.Latitude, 
                            lastFetchMax.Latitude);
                } else {
                    fetcher = new TrafikInfoNuCountyFetcher(deviceId, versionCode, regionId);
                }
                
                fetcher.SetStatusReciever(new IStatusReciever() {
                    @Override
                    public void RecieveStatus(Status status) {
                        String statusString = getStatusString(status);
                        reportStatus(statusString, 0, 0);
                    }

                    @Override
                    public void RecieveStatus(Status status, int done, int total) {
                        String statusString = getStatusString(status);
                        statusString += " (" + done + " av " + total + ")";
                        reportStatus(statusString, done, total);
                    }

                    private String getStatusString(Status status) {
                        // TODO Hardcoded strings
                        String statusString = "";
                        if (status == Status.Downloading)
                            statusString += "Hämtar data";
                        else if(status == Status.Extracting)
                            statusString += "Behandlar data";
                        else if (status == Status.Parsing)
                            statusString += "Läser medddelande";
                        return statusString;
                    }
                });

                SituationHandler handler = new SituationHandler(
                        mDbAdapter, 
                        prefHandler.getCategoryMask(getBaseContext()));
                
                fetcher.getSituations(lastUpdateTimestamp, handler);
                
                List<String> changedIds = handler.getChangedIds();

                lastUpdateTimestamp = fetcher.getLastTimestamp();
                mDbAdapter.setServerTimestamp(lastUpdateTimestamp);
                
                // TODO do not clean up every minute
                Date now = new Date();
                Date lastTimeStamp = mDbAdapter.getClenupTimestamp();
                Long cleanUpIntervall = 1000 * 60 * 60 * 3L; // 3 hours
                if (  regionChanged ||
                      lastUpdateTimestamp == null || 
                      lastTimeStamp == null || 
                      (now.getTime() - lastTimeStamp.getTime() > cleanUpIntervall)){
                    Log.d(LOGTAG,"Time to clean up obsolete data");
                    cleanUpData(fetcher);
                    mDbAdapter.setCleanupTimestamp(now);
                }

                reportStatus("Uppdatering klar", 0, 0);

                mDbAdapter.setUpdateTimestamp(lastCompleteUpdate);
                
                Boolean dataChanged = !changedIds.isEmpty();
                updateUserNotification(InfoFetchService.this, mDbAdapter, dataChanged);

                if (msg.what == MESSAGE_COMPLETE)
                    reportCompleteUpdateDone();

            } catch (FetchInfoException e) {
                reportError(e);
            } catch (SQLException e) {
                e.printStackTrace();
                Log.e(LOGTAG,"SQLException in handleMessage: " + e.getMessage());
                // TODO: Report DB error
                reportError(new FetchInfoException("Database", ErrorCodes.DatabaseError));
            } catch (RuntimeException e) {
             // TODO remove this catch block
                Log.e(LOGTAG,"RuntimeException in handleMessage: " + e.getMessage());
                throw e;
            } finally {
                try {
                    mDbAdapter.close();

                    Date lastStartTime = mLastStartTime;
                    if (lastStartTime != null
                            && lastStartTime.compareTo(lastCompleteUpdate) <= 0) {
                        Log.d(LOGTAG, "Calling stopSelf");
                        mLastStartTime = null;
                        
                        stopSelf();                    

                        PowerManager.WakeLock lock = getLock(InfoFetchService.this);
                        if (lock.isHeld())
                            lock.release();
                    }
                } catch (RuntimeException e) {
                    // TODO remove this catch block
                    Log.e(LOGTAG,"RuntimeException in handleMessage.finally: " + e.getMessage());
                    throw e;
                }
            }
        }

        private void cleanUpData(IDataFetcher fetcher) throws FetchInfoException {
            fetcher.SetStatusReciever(null);
            reportStatus("Kontrollerar föråldrarde meddelanden", 0, 0);
            List<String> ids = fetcher.getAllIds();

            List<Long> test = new ArrayList<Long>(ids.size());
            for (String sid : ids){
                test.add(mDbAdapter.getIdByExternalId(sid));
            }
            
            reportStatus("Raderar gamla meddelanden", 0, 0);

            mDbAdapter.markInformatiomObsoleteNotInList(test);

            // TODO hardcoded number of days to save obsolete situations
            long ms = (new Date()).getTime();
            ms -= 1000 * 60 * 60 * 24 * 14; // 14 days
            Date deleteOlderThan = new Date(ms);
            mDbAdapter.deleteOldObsoleteSituations(deleteOlderThan);
        }

        private Position getCurrentPosition(int minNetworkAccurency) {
            // TODO Create and use preferences?
            reportStatus("Söker aktuell position", 0, 0);

            PreferenceHandler prefHandler = new PreferenceHandler();
            final boolean useGps = prefHandler.getUseGps(InfoFetchService.this);
            LocationRetriever locRet = new LocationRetriever(useGps, 15000, 60 * 1000, minNetworkAccurency);
            Location currentLocation = locRet.getCurrentLocation(InfoFetchService.this);
            Position position;
            if (currentLocation == null){
                reportStatus("Ingen position funnen", 0, 0);
                position = mDbAdapter.getPoint(InfoDbAdapter.POSITIONTYPE_LASTPOSITION, new Position());
            } else {
                reportStatus("Aktuell position funnen", 0, 0);
                position = new Position();
                position.Latitude = currentLocation.getLatitude();
                position.Longitude = currentLocation.getLongitude();
                mDbAdapter.setPoint(InfoDbAdapter.POSITIONTYPE_LASTPOSITION, position);
            }
            return position;
        }

       
        private int getRegionIdFromPreference() {
            Log.d(LOGTAG, "In getRegionIdFromPreference");
            SharedPreferences prefs = PreferenceManager
                    .getDefaultSharedPreferences(InfoFetchService.this);
            
            String regionPreferenceKey = getResources().getString(
                    R.string.regionKey);
            
            String defaultRegionId = getResources().getString(
                    R.string.DefaultRegionId);
            
            String regionIdString = prefs.getString(regionPreferenceKey,
                    defaultRegionId);
            
            Log.d(LOGTAG, "regionid = " + regionIdString);
            
            if (regionIdString.contains(",")){
                regionIdString = translateOldPreference(regionIdString, defaultRegionId);
                Editor editor = prefs.edit();
                editor.putString(regionPreferenceKey, regionIdString);
                editor.commit();
                Log.d(LOGTAG, "regionid  changed to " + regionIdString);
            }
            
            int id = Integer.parseInt(regionIdString);           
            
            return id;
        }

        private String translateOldPreference(String oldString, String defualtId) {
            String newId = defualtId;
            if (oldString.contentEquals("4359,4356"))
                newId = "8"; // Kalmar
            else if (oldString.contentEquals("4350,4347"))
                newId = "12"; // Skåne
            else if (oldString.contentEquals("4350,4347"))
                newId =  "2"; // Stockholm
            else if (oldString.contentEquals(" 1234,1234"))
                newId = "14"; // Västra götaland
            
            return newId;
        }

        private void reportError(FetchInfoException e) {
            Log.d(LOGTAG, "In reportError");
            for (IDataObserver observer : mCallbacks) {
                observer.errorEncountered(e);
            }
            Bundle extras = new Bundle();
            extras.putSerializable(UPDATE_INTENT_EXTRA_EXCEPTION, e);

            
            sendUpdateBroadcastIntent(extras);
        }

        private void reportCompleteUpdateDone() {
            Log.d(LOGTAG, "In reportCompleteUpdateDone");
            for (IDataObserver observer : mCallbacks) {
                observer.completeDataUpdateDone();
            }
            
            sendUpdateBroadcastIntent();
        }

        private void reportStatus(String status, int done, int total) {
            Log.i(LOGTAG, "Status: " + status);
            for (IDataObserver observer : mCallbacks) {
                observer.setStatus(status, done, total);
            }
            
            Bundle extras = new Bundle();
            extras.putString(UPDATE_INTENT_EXTRA_STATUS, status);
            extras.putInt(UPDATE_INTENT_EXTRA_DONE, done);
            extras.putInt(UPDATE_INTENT_EXTRA_TOTAL, total);
            
            sendUpdateBroadcastIntent(extras);
        }

        private void sendUpdateBroadcastIntent() {
            sendUpdateBroadcastIntent(null);
        }

        private void sendUpdateBroadcastIntent(Bundle extras) {
            Context context = getBaseContext();
            sendUpdateIntent(extras, context);
        }
    }
    
    public static void updateUserNotification(Context context, InfoDbAdapter dbAdapter, Boolean dataChanged){
        Log.d(LOGTAG, "in updateUserNotification");
        Log.d(LOGTAG, "in updateUserNotification datChanged = " + dataChanged);
        
        // TODO Hardcoded strings
        NotificationManager notificationManager = 
            (NotificationManager) context.getSystemService(Context.NOTIFICATION_SERVICE);
        
        PreferenceHandler prefHandler = new PreferenceHandler();
        byte preferredNotificationLevel = prefHandler.getNoticifationLevel(context);
        
        if (preferredNotificationLevel == PreferenceHandler.NOTIFICATIONLEVEL_NOTHING){
            Log.d(LOGTAG, "in updateUserNotification no notification requested");
            notificationManager.cancel(NOTIFICATION_ID);
            return;
        }
        
        int icon = R.drawable.icon;
        CharSequence tickerText = "Olästa trafikmeddelanden";
        long when = System.currentTimeMillis();

        Notification notification = new Notification(icon, tickerText, when);
        
        int numberOfUnread = dbAdapter.getNumberOfUnreadSituations();
        
        if (numberOfUnread == 0){
            Log.d(LOGTAG, "in updateUserNotification no unread situations. cancel notification.");
            notificationManager.cancel(NOTIFICATION_ID);
        } else {
            String contentTitle;
            String contentText;
            Intent notificationIntent;
            
            if (numberOfUnread == 1){
                Log.d(LOGTAG, "in updateUserNotification one unread situation.");
                TrafficInfo info = dbAdapter.getNewest();
                
                contentTitle = info.getHeaderWithRoadname();
                
                contentText = info.getTypeName();
                
                notificationIntent = new Intent(context, DetailedActivity.class);
                notificationIntent.putExtra(DetailedInfoActivity.INTENTEXTRA_INFO, info);
                notificationIntent.putExtra(DetailedInfoActivity.INTENTEXTRA_MARKREAD, true);
            } else {
                Log.d(LOGTAG, "in updateUserNotification multiple unread situations.");
                contentTitle = numberOfUnread + " olästa trafikmeddelanden";
                contentText = "";
                notificationIntent = new Intent(context, OverviewActivity.class);    
            }

            if (dataChanged && preferredNotificationLevel >= PreferenceHandler.NOTIFICATIONLEVEL_ICONANDSOUND){
                Log.d(LOGTAG, "in updateUserNotification add sound to notification.");
                notification.defaults |= Notification.DEFAULT_SOUND;
            }
            
            if (dataChanged && preferredNotificationLevel >= PreferenceHandler.NOTIFICATIONLEVEL_ICONSOUNDANDLED){
                // notification.defaults |= Notification.DEFAULT_LIGHTS;
                Log.d(LOGTAG, "in updateUserNotification add LED to notification.");
                notification.ledARGB = 0xFF0000FF;
                notification.flags = Notification.FLAG_SHOW_LIGHTS;
                notification.ledOnMS = 200;
                notification.ledOffMS = 3000;
            }
 
            
            PendingIntent contentIntent = PendingIntent.getActivity(context, 0, notificationIntent, PendingIntent.FLAG_UPDATE_CURRENT);

            notification.setLatestEventInfo(context, contentTitle, contentText, contentIntent);
            
            notificationManager.notify(NOTIFICATION_ID, notification);                
        }
    }

    public static void sendUpdateIntent(Bundle extras, Context context) {
        Intent updateIntent = new Intent(UPDATE_INTENT);
        
        if (extras != null)
            updateIntent.putExtras(extras);

        PendingIntent pendingupdateIntent = PendingIntent.getBroadcast(
                context, 0, updateIntent,
                PendingIntent.FLAG_UPDATE_CURRENT);
        try {
            pendingupdateIntent.send();
        } catch (CanceledException e) {
            Log.w(LOGTAG, "CanceledException when sending intent");
        }
    }
    
    public void addDataObserver(IDataObserver observer) {
        Log.d(LOGTAG, "In addDataObserver");
        if (!mCallbacks.contains(observer)) {
            Log.d(LOGTAG, "DataObserver added");
            mCallbacks.add(observer);
        }
    }

    public void removeDataObserver(IDataObserver oberver) {
        Log.d(LOGTAG, "DataObserver removed");
        mCallbacks.remove(oberver);
    };

}
