/*
    This source file is part of the trafikinfo for android project
    Copyright (C) 2010  Peter Björkman


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package se.peterbjorkman.android.trafikinfo;

import java.util.Date;

import se.peterbjorkman.android.trafikinfo.TrafficInfo.ReadState;

import android.database.Cursor;
import android.text.Html;
import android.view.View;
import android.widget.ImageView;
import android.widget.SimpleCursorAdapter;
import android.widget.TextView;

public class ListViewBinder implements SimpleCursorAdapter.ViewBinder {
    private static final String LOGTAG = Logger.createLogtag(ListViewBinder.class);
    
    private DateFormatter mDateFormatter;

    public ListViewBinder() {
        Logger.d(LOGTAG, "In constructor");
        mDateFormatter = new DateFormatter("yyyy-MM-dd", "HH:mm");
    }

    @Override
    public boolean setViewValue(View view, Cursor cursor, int columnIndex) {
        Logger.d(LOGTAG, "In setViewValue");
        String name = cursor.getColumnName(columnIndex);
        if (name.contentEquals(InfoDbAdapter.KEY_FROM)) {
            bindPeriodView(view, cursor, columnIndex);
        } else if (name.contentEquals(InfoDbAdapter.KEY_LOCATION)
                || name.contentEquals(InfoDbAdapter.KEY_TYPENAME)) {
            bindOtherView(view, cursor, columnIndex);
        }
        else if (name.contentEquals(InfoDbAdapter.KEY_HEADER)) {
            bindHeaderView(view, cursor, columnIndex);
        } else if (name.contentEquals(InfoDbAdapter.KEY_TYPE)) {
            bindTypeView(view, cursor, columnIndex);
        }

        return true;
    }

    private void bindTypeView(View view, Cursor cursor, int columnIndex) {
        Logger.d(LOGTAG, "In setViewValue name is KEY_TYPE");
        int type = cursor.getInt(cursor.getColumnIndexOrThrow(InfoDbAdapter.KEY_TYPE));
        int imageId = TrafficInfo.getIconResourceByType(type);
        ImageView imageView = (ImageView)view;
        imageView.setImageResource(imageId);
    }

    private void bindOtherView(View view, Cursor cursor, int columnIndex) {
        Logger.d(LOGTAG, "In setViewValue name is KEY_LOCATION or KEY_TYPENAME");
        String value = cursor.getString(columnIndex);
        if (value.length() == 0){
            Logger.d(LOGTAG, "In setViewValue hiding view");
            view.setVisibility(View.GONE);
        }
        else{
            Logger.d(LOGTAG, "In setViewValue showing view");
            
            ReadState state = extractReadState(cursor);
            setText(view, value, state);
        }
    }

    private void bindHeaderView(View view, Cursor cursor, int columnIndex) {
        Logger.d(LOGTAG, "In bindHeaderView");
        String header = cursor.getString(columnIndex);
        String roadname = cursor.getString(cursor
                .getColumnIndexOrThrow(InfoDbAdapter.KEY_ROAD));
        
        header = TrafficInfo.getHeaderWithRoadName(header, roadname);
        
        ReadState state = extractReadState(cursor);
        
        // TODO hardcoded strings
        if (state == ReadState.New)
            header = "(Ny) " + header;
        
        if (state == ReadState.Modified)
            header = "(Uppdaterad) " + header;
        
        setText(view, header, state);
    }

    private void bindPeriodView(View view, Cursor cursor, int columnIndex) {
        Logger.d(LOGTAG, "In bindPeriodView");
        
        Date from = extractDate(cursor, columnIndex);
        
        Date to = extractDate(cursor, cursor
                .getColumnIndexOrThrow(InfoDbAdapter.KEY_TO));
        
        String periodText = mDateFormatter.FormatPeriod(from, to);
        
        ReadState state = extractReadState(cursor);
        setText(view, periodText, state);
    }
    
    private Date extractDate(Cursor curs, int columnIndex) {
        Logger.d(LOGTAG, "In extractDate");
        Date date;
        if (curs.isNull(columnIndex))
            date = null;
        else
            date = new Date(curs.getLong(columnIndex));
        return date;
    }

    private ReadState extractReadState(Cursor cursor) {
        Logger.d(LOGTAG, "In extractReadState");
        int stateint = cursor.getInt(cursor.getColumnIndexOrThrow(InfoDbAdapter.KEY_READ));
        return ReadState.values()[stateint];
    }

    private void setText(View view, String text, ReadState state) {
        Logger.d(LOGTAG, "In setText");
        TextView tv = (TextView)view;
        
        if (state == ReadState.New || state == ReadState.Modified)
            tv.setText(Html.fromHtml("<b>" + text + "</b>"));
        else
            tv.setText(text);
        
        view.setVisibility(View.VISIBLE);
    }
}
