package se.peterbjorkman.android.trafikinfo;

import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoException;
import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.os.Handler;
import android.os.IBinder;
import android.os.Message;
import android.util.Log;

public class ServiceBinder implements IDataObserver {
    private static final String LOGTAG = ServiceBinder.class.getName();
    
    private static final int MESSAGE_SET_STATUS = 1;
    private static final int MESSAGE_DATA_UPDATE = 2;
    private static final int MESSAGE_ERROR_ENCOUNTERED = 4;

    private static final String MESSAGE_SET_STATUS_STATUS = "status";
    private static final String MESSAGE_SET_STATUS_TOTAL = "total";
    private static final String MESSAGE_SET_STATUS_DONE = "done";

    private static final String MESSAGE_ERROR_ENCOUNTERED_EXCEPTION = "exception";

    private InfoFetchService mBoundService;
    
    private final IDataObserver mObserver;
    
    public ServiceBinder(IDataObserver observer){
        mObserver = observer;
    }
    
    public void bind(Context context){
        Boolean success = context.bindService(new Intent(context, InfoFetchService.class),
                mConnection, Context.BIND_AUTO_CREATE);
        
        if (!success)
            Log.w(LOGTAG, "Binding to service returned false.");
    }
    
    public void unbind(Context context){
        if (mBoundService != null) {
            mBoundService.removeDataObserver(this);
            context.unbindService(mConnection);
        }
    }
    
    private ServiceConnection mConnection = new ServiceConnection() {
        public void onServiceConnected(ComponentName className, IBinder service) {

            Log.d(LOGTAG, "In onServiceConnected");
            
            // This is called when the connection with the service has been
            // established, giving us the service object we can use to
            // interact with the service. Because we have bound to a explicit
            // service that we know is running in our own process, we can
            // cast its IBinder to a concrete class and directly access it.
            mBoundService = ((InfoFetchService.LocalBinder) service)
                    .getService();
            
            mBoundService.addDataObserver(ServiceBinder.this);
        }

        public void onServiceDisconnected(ComponentName className) {
            
            Log.d(LOGTAG, "In onServiceDisconnected");
            
            // This is called when the connection with the service has been
            // unexpectedly disconnected -- that is, its process crashed.
            // Because it is running in our same process, we should never
            // see this happen.
            mBoundService = null;
        }
    };
    
    /** Message handler for messages from background threads. */
    Handler viewUpdateHandler = new Handler() {
        @Override
        public void handleMessage(Message msg) {
            
            Log.d(LOGTAG, "In handleMessage msg.what = " + msg.what);
            
            switch (msg.what) {

            case MESSAGE_SET_STATUS:
                String status = msg.getData().getString(
                        MESSAGE_SET_STATUS_STATUS);
                int done = msg.getData().getInt(MESSAGE_SET_STATUS_DONE);
                int total = msg.getData().getInt(MESSAGE_SET_STATUS_TOTAL);
                mObserver.setStatus(status, done, total);
                return;

            case MESSAGE_DATA_UPDATE:
                mObserver.completeDataUpdateDone();
                return;

            case MESSAGE_ERROR_ENCOUNTERED:
                FetchInfoException e = (FetchInfoException) msg.getData()
                        .getSerializable(MESSAGE_ERROR_ENCOUNTERED_EXCEPTION);
                mObserver.errorEncountered(e);
                return;

            }
            Log.d(LOGTAG, "Message not handled in handleMessage. msg.what = " + msg.what);
            super.handleMessage(msg);
        }
    };

    @Override
    public void completeDataUpdateDone() {
        Log.d(LOGTAG, "in completeDataUpdateDone");
        Message m = new Message();
        m.what = MESSAGE_DATA_UPDATE;
        viewUpdateHandler.sendMessage(m);
    }

    @Override
    public void setStatus(String status, int done, int total) {
        Log.d(LOGTAG, "in setStatus");
        Message m = new Message();
        m.what = MESSAGE_SET_STATUS;
        m.getData().putInt(MESSAGE_SET_STATUS_DONE, done);
        m.getData().putInt(MESSAGE_SET_STATUS_TOTAL, total);
        m.getData().putString(MESSAGE_SET_STATUS_STATUS, status);
        viewUpdateHandler.sendMessage(m);
    }

    @Override
    public void errorEncountered(FetchInfoException e) {
        Log.d(LOGTAG, "in errorEncountered");
        Message m = new Message();
        m.what = MESSAGE_ERROR_ENCOUNTERED;
        m.getData().putSerializable(MESSAGE_ERROR_ENCOUNTERED_EXCEPTION, e);
        viewUpdateHandler.sendMessage(m);
    }

}
