/*
    This source file is part of the trafikinfo for android project
    Copyright (C) 2009-2010  Peter Björkman


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package se.peterbjorkman.android.trafikinfo;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoException;
import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoHttpException;
import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoException.ErrorCodes;
import android.app.ListActivity;
import android.content.Intent;
import android.database.Cursor;
import android.database.SQLException;
import android.os.Bundle;
import android.util.Log;
import android.view.ContextMenu;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.ContextMenu.ContextMenuInfo;
import android.widget.ListAdapter;
import android.widget.ListView;
import android.widget.ProgressBar;
import android.widget.SimpleCursorAdapter;
import android.widget.TextView;
import android.widget.AdapterView.AdapterContextMenuInfo;

public abstract class SituationListActivity extends ListActivity implements IDataObserver {
    private static final String LOGTAG = SituationListActivity.class.getName();
    private static final int CONTEXTMENU_DELETE_ID = Menu.FIRST + 100;
    
    private InfoDbAdapter mDbAdapter;
    private DateFormatter formatter = new DateFormatter("yyyy-MM-dd", "HH:mm");
    private ServiceBinder mBinder = new ServiceBinder(this);
    
    private List<Long> mIdList;

    @Override
    public void onCreate(Bundle savedInstanceState) {
        Log.d(LOGTAG, "In onCreate");
        super.onCreate(savedInstanceState);
        setContentView(R.layout.overview);
        
        mBinder.bind(getApplicationContext());
        
        registerForContextMenu(getListView());
        
        try {
            mDbAdapter = new InfoDbAdapter(this);
            mDbAdapter.open();
        } catch (SQLException e) {
            e.printStackTrace();
            Log.e(LOGTAG, "SQLException: " + e.getLocalizedMessage());
            errorEncountered(new FetchInfoException("Database", ErrorCodes.DatabaseError));
        }
        Log.d(LOGTAG, "Leaving onCreate");
    }

    protected abstract Cursor getCursor() ;
    
    protected InfoDbAdapter getDbAdapter(){
        return mDbAdapter;
    }
    
    @Override
    public void onResume() {
        super.onResume();
        InfoFetchService.updateUserNotification(this, mDbAdapter, false);
        requery();
    }
    
    @Override
    public void onStop() {
        // HACK: Could not get anything from the listadapter in onDestroy. 
        // Get it here and save it to use it in onDestroy
        mIdList = getIdListFromAdapter();
        super.onStop();
    }

    @Override
    protected void onDestroy() {
        Log.d(LOGTAG, "In onDestroy");

        mBinder.unbind(getApplicationContext());
        markAllAsRead(mIdList);
        mDbAdapter.close();
        super.onDestroy();
        Log.d(LOGTAG, "Leaving onDestroy");
    }

    private void markAllAsRead(List<Long> ids) {
        Log.d(LOGTAG, "in markAllAsRead");
        if (ids == null)
            Log.w(LOGTAG, "List is null in markAllAsRead");
        else
            mDbAdapter.markSiuationsAsRead(ids, this);
    }

    private List<Long> getIdListFromAdapter() {
        Log.d(LOGTAG, "in getIdListFromAdapter");
        ListAdapter adapter = getListAdapter();
        
        if (adapter == null)
            return null;
        
        int numberOfSituations = adapter.getCount();

        List<Long> ids = new ArrayList<Long>(numberOfSituations);
        for (int i = 0; i < numberOfSituations ; i++ ){
            long id = adapter.getItemId(i);
            ids.add(id);
        }
        return ids;
    }
    
    @Override
    protected void onListItemClick(ListView l, View v, int position, long id) {
        Log.d(LOGTAG, "In onListItemClick position = " + position);
        Intent launchIntent = new Intent(SituationListActivity.this,
                DetailedActivity.class);
        TrafficInfo obj = mDbAdapter.getInformationById(id);
        
        List<Long> ids = new ArrayList<Long>(1);
        ids.add(id);
        mDbAdapter.markSiuationsAsRead(ids, this);
        
        launchIntent.putExtra(DetailedInfoActivity.INTENTEXTRA_INFO, obj);
        Log.d(LOGTAG, "Starting DetailedActivity");
        SituationListActivity.this.startActivity(launchIntent);
        Log.d(LOGTAG, "after intent to DetailedActivity");
    }

    @Override
    public boolean onContextItemSelected(MenuItem item)
    {
        switch (item.getItemId()) {
        case CONTEXTMENU_DELETE_ID:
          AdapterContextMenuInfo info = (AdapterContextMenuInfo) item.getMenuInfo();

          mDbAdapter.markSiuationDeleted(info.id, getBaseContext());
          requery();
          return true;
        default:
          return super.onContextItemSelected(item);
        }
    }

    @Override
    public void onCreateContextMenu(ContextMenu menu, View v, ContextMenuInfo menuInfo)
    {
        menu.add(ContextMenu.NONE, CONTEXTMENU_DELETE_ID,ContextMenu.NONE, R.string.deleteMessage);
    }

    protected void requery() {
        Log.d(LOGTAG, "In requery");
        try{
            Cursor notesCursor = getCursor();
    
            startManagingCursor(notesCursor);
            
            String[] columns = new String[] { InfoDbAdapter.KEY_HEADER,
                    InfoDbAdapter.KEY_FROM, InfoDbAdapter.KEY_LOCATION,
                    InfoDbAdapter.KEY_TYPENAME, InfoDbAdapter.KEY_TYPE };
    
            int[] views = new int[] { R.id.info_row_description,
                    R.id.info_row_period, R.id.info_row_position,
                    R.id.info_row_type, R.id.info_row_image};
    
            SimpleCursorAdapter notes = new SimpleCursorAdapter(this,
                    R.layout.info_row, notesCursor, columns, views);
    
            notes.setViewBinder(new ListViewBinder());
    
            setListAdapter(notes);
            SetUpdateTimestampText();
        } catch (SQLException e) {
            e.printStackTrace();
            Log.e(LOGTAG, "SQLException: " + e.getLocalizedMessage());
            errorEncountered(new FetchInfoException("Database", ErrorCodes.DatabaseError));
        }
        Log.d(LOGTAG, "Leavng requery");
    }

    private void SetUpdateTimestampText() {
        Log.d(LOGTAG, "In SetUpdateTimestampText");
        hideProgressBar();

        TextView textView = (TextView) findViewById(R.id.ProgressTextView);

        Date timestamp = mDbAdapter.getUpdateTimestamp();
        if (timestamp != null) {
            String status = getResources().getString(
                    R.string.status_list_updated)
                    + " ";
            status += formatter.FormatTimestamp(timestamp);
            textView.setText(status);
        } else {
            Log.d(LOGTAG, "Timestamp not set. Set text to empty string");
            textView.setText("");
        }
    }

    /**
     * Hides the rotating progress indicator.
     * <p>
     * Should only be called by the GUI-thread.
     */
    private void hideProgressBar() {
        Log.d(LOGTAG, "In hideProgressBar");
        View progressbar = findViewById(R.id.ProgressBar);
        progressbar.setVisibility(View.GONE);
    }

    /**
     * Writes a status string to the status field and sets the progress of the
     * rotating progress indicator.
     * <p>
     * Should only be called by the GUI-thread.
     * 
     * @param status
     *            Text that should be written to the stus field
     * @param done
     *            How much of the total work that is completed
     * @param total
     *            A value indicating how much work is to be done. If total is
     *            zero the rotating progress bar is set to indeterminate.
     */
    @Override
    public void setStatus(String status, int done, int total) {
        Log.d(LOGTAG, "In setStatusUnsafe");
        View progressbar = findViewById(R.id.ProgressBar);
        progressbar.setVisibility(View.VISIBLE);
        TextView textView = (TextView) findViewById(R.id.ProgressTextView);
        textView.setText(status);
        ProgressBar progressBar = (ProgressBar) findViewById(R.id.ProgressBar);
        if (total == 0){
            Log.d(LOGTAG, "set progressbar to indeterminate");
            progressBar.setIndeterminate(true);
        }
        else {
            progressBar.setIndeterminate(false);
            progressBar.setMax(total);
            progressBar.setProgress(done);
        }
    }

    /**
     * Displays an exception to the user to indicate something went wrong.
     * <p>
     * Should only be called by the GUI-thread.
     */
    @Override
    public void errorEncountered(FetchInfoException e) {
        Log.d(LOGTAG, "In handleException");
        String errorString = getResources().getString(R.string.Error_title)
                + "\n";
        errorString += getResources().getString(e.getStringID()) + "\n";
        if (e.getErrorCode() == ErrorCodes.HttpError) {
            FetchInfoHttpException httpException = (FetchInfoHttpException) e;
            errorString += getResources().getString(R.string.Error_code) + ":";
            errorString += httpException.getHttpErrorCode();
        }
        // errorString += getResources().getString(R.string.adress) + ":\n";
        // errorString += e.getUrl();
        hideProgressBar();

        TextView textView = (TextView) findViewById(R.id.ProgressTextView);
        textView.setText(errorString);

        // TODO: show the error message in some other way
    }

    @Override
    public void completeDataUpdateDone() {
        requery();
    }
}
