package se.peterbjorkman.android.trafikinfo;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.HttpConnectionParams;
import org.apache.http.params.HttpParams;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import android.os.Build;
import android.util.Log;

import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoException;
import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoHttpException;
import se.peterbjorkman.android.trafikinfo.Exceptions.FetchInfoException.ErrorCodes;
import se.peterbjorkman.android.trafikinfo.IStatusReciever.Status;
import se.peterbjorkman.android.trafikinfo.TrafficInfo.ReadState;

public abstract class TrafikInfoNuInfoFetcher implements IDataFetcher {

    private static final String LOGTAG = TrafikInfoNuInfoFetcher.class
            .getName();
    
    private final String mDeviceId;
    private final int mVersionCode;
    private final SimpleDateFormat parse = new SimpleDateFormat("yyyy-MM-dd HH:mm");
    private final SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd+HH:mm");
    private Date mLastTimestamp = null;
    private IStatusReciever mStatusReciver = null;
    
    
    public TrafikInfoNuInfoFetcher(String deviceId, int versionCode)
    {
        mDeviceId = deviceId;
        mVersionCode = versionCode;
    }

    @Override
    public TrafficInfo GetMessageById(int id) throws FetchInfoException {
        List<Integer> ids = new ArrayList<Integer>(1);
        ids.add(id);
        List<TrafficInfo> infos = getSituations(ids);
        
        int numberOfSituationsFound = infos.size();
        if (numberOfSituationsFound == 1)
            return infos.get(0);
        
        // TODO throw exception if zero or multiple situations retrived
        return null;
    }

    @Override
    public List<String> getAllIds() throws FetchInfoException {
        return getIdsNewerThan(null);
    }

    protected List<String> getIdsByUrl(Date onlyNewerThan, String url)
            throws FetchInfoException {
        url += "&uid=" + mDeviceId;
        url += "&ver=" + mVersionCode;
        url += "&os=" + Build.VERSION.SDK; // TODO Use SDKINT When skipping 1.5
        
        if (onlyNewerThan != null)
            url += "&modifiedSince=" + formatter.format(onlyNewerThan);
        
        try {
            JSONObject json = DownloadJSon(url);
            setStatus(Status.Extracting);
            // String timestamp = json.getString("timeStamp"); // TODO return timestamp somehow
            
            verifyApiVersion(json, url);

            JSONArray situations = json.getJSONArray("situations");
            int count = situations.length();
            List<String> ids = new ArrayList<String>(count);
            for (int i = 0; i<count;i++){
                String id = situations.getString(i);
                ids.add(id);
            }
            return ids;
        } catch (JSONException e) {
            Log.e(LOGTAG, "JSONException in getIdsNewerThan.");
            Log.e(LOGTAG, "JSONException message: " + e.getMessage());
            throw new FetchInfoException(url,
                    ErrorCodes.DataFormatError, e);
        }
    }

    private void verifyApiVersion(JSONObject json, String url) throws FetchInfoException, JSONException {
        int compApi = json.getInt("compapi");
        if (compApi > 1)
            throw new FetchInfoException(url, ErrorCodes.ServerApiVersionError);   
    }

    private List<TrafficInfo> getSituations(List<Integer> ids) throws FetchInfoException {
        
        int count = ids.size();
        List<TrafficInfo> ret = new ArrayList<TrafficInfo>();
        
        if (count  < 1)
            return ret;
        
        String url = "http://api.trafikinfo.nu/getSituations.php?lang=sv-SE&situations=";
        url = url + ids.get(0);
        
        for (int i=1; i < count;i++){
            url = url + "," + ids.get(i);
        }
        
        try {
            
            JSONObject obj;
            obj = DownloadJSon(url);

            JSONArray situations = obj.getJSONArray("situations");
            int situationsCount = situations.length();
            for (int i = 0 ; i < situationsCount; i++){
                setStatus(Status.Extracting, i, count);
               JSONObject situation = situations.getJSONObject(i);
               TrafficInfo info = extractTrafficInfoFromJson(situation);
               ret.add(info);
            }
            return ret;
        } catch (JSONException e) {
            Log.e(LOGTAG, "JSONException in getSituations.");
            Log.e(LOGTAG, "JSONException message: " + e.getMessage());
            throw new FetchInfoException(url,
                    ErrorCodes.DataFormatError, e);
        }
    }

    private TrafficInfo extractTrafficInfoFromJson(JSONObject situation) throws JSONException {
        String id = situation.getString("id");
        String header = situation.optString("header");
        String location = situation.optString("location");

        List<String> commentList = extractComments(situation);
        
        JSONArray restrictionsJson = situation.getJSONArray("resrictions");
        int numberOfRestrictions = restrictionsJson.length();
        List<Restriction> retrictionList = new ArrayList<Restriction>(numberOfRestrictions);
        
        for (int i= 0; i < numberOfRestrictions; i++)
        {
            JSONObject restrictionJson = restrictionsJson.getJSONObject(i);
            Restriction restriction = extractRestritionFromJson(restrictionJson, i);
            retrictionList.add(restriction);
        }
        
        int type = situation.getInt("type");
        String typeName = situation.getString("typename");
        
        // == null: Work around for server or SRA-bug
        if (header.contentEquals("null") || header.length() == 0)
            header = typeName;
        
        String source = situation.optString("source"); 
        String road = situation.optString("road");
        Date modified = extractDate(situation, "modified");;
        Double longitude = extractDouble(situation,"longitude");
        Double latitude = extractDouble(situation,"latitude");
        Date to = extractDate(situation, "to");
        Date from = extractDate(situation, "from");

        TrafficInfo info = new TrafficInfo(id, 
                header, 
                type, 
                CategoryResolver.getCategoryByType(type),
                typeName,
                source, 
                location,
                road,
                modified,
                longitude,
                latitude,
                from, 
                to,
                ReadState.Unknown,
                commentList,
                retrictionList);
        return info;
    }

    private List<String> extractComments(JSONObject situation)
            throws JSONException {
        JSONArray comments = situation.getJSONArray("comments");
        int numberOfComments = comments.length();
        List<String> commentList = new ArrayList<String>(numberOfComments);
        
        for (int i= 0; i < numberOfComments; i++)
        {
            String comment = comments.getString(i);
            commentList.add(comment);
        }
        return commentList;
    }

    private Restriction extractRestritionFromJson(JSONObject restrictionJson, int number) throws JSONException {
        int type = restrictionJson.getInt("type");
        String typeName = restrictionJson.getString("typename");
        int speedLimit = restrictionJson.optInt("speedLimit",0); // TODO: NULL? 
        Date to = extractDate(restrictionJson, "to");
        Date from = extractDate(restrictionJson, "from");
        List<String> commentList = extractComments(restrictionJson);
        Restriction restriction = new Restriction(
                from, 
                to, 
                number, 
                type, 
                typeName, 
                speedLimit, 
                commentList);
        return restriction;
        }

    private Double extractDouble(JSONObject situation, String name) {
        String stringVal = situation.optString(name);
        Double val = null;
        if (!stringVal.contentEquals(""))
            try {
                val = Double.parseDouble(stringVal);
            } catch (NumberFormatException e) {
                val = null;             
            }
        return val;
    }

    private Date extractDate(JSONObject situation, String name) {
        String toDateString = situation.optString(name);
        Date to = null;
        if (!toDateString.contentEquals(""))
            try {
                to = parse.parse(toDateString);
            } catch (ParseException e) {
                to = null;             
            }
        return to;
    }

    protected JSONObject DownloadJSon(String url) throws FetchInfoException, JSONException {
        Log.d(LOGTAG, "In DownloadJSon url = " + url);
        String jsonString = null;
        try {
            final int timeoutSocket = 60000;
            final int timeoutConnection = 15000;
            
            setStatus(Status.Downloading);
            
            HttpGet request = new HttpGet(url);
            HttpParams httpParameters = new BasicHttpParams();
            HttpConnectionParams.setConnectionTimeout(httpParameters, timeoutConnection);
            HttpConnectionParams.setSoTimeout(httpParameters, timeoutSocket);
            
            HttpClient client = new DefaultHttpClient(httpParameters);
            HttpResponse response = client.execute(request);

            int status = response.getStatusLine().getStatusCode();

            if (status != HttpStatus.SC_OK) {
                String str = getHttpResonseString(response).trim();
                if (str.length() > 100)
                    str = "";
                throw new FetchInfoHttpException(url, status, str);
            }

            ByteArrayOutputStream outStream = new ByteArrayOutputStream();

            response.getEntity().writeTo(outStream);

            jsonString = outStream.toString();
            setStatus(Status.Parsing);
            return new JSONObject(jsonString);

        } catch (IOException e1) {
            Log.e(LOGTAG, "IOException in DownloadJSon.");
            Log.e(LOGTAG, "Url: " + url);
            Log.e(LOGTAG, "IOException message: " + e1.getMessage());
            throw new FetchInfoException(url,
                    ErrorCodes.IOError, e1);
        }
    }

    private String getHttpResonseString(HttpResponse response)
            throws IOException, UnsupportedEncodingException {
        Log.d(LOGTAG, "In getHttpResonseString");
        InputStream content = response.getEntity().getContent();

        StringBuilder sb = new StringBuilder();
        String line;

        try {
            Header head = response.getEntity().getContentEncoding();
            BufferedReader reader;

            if (head != null)
                reader = new BufferedReader(new InputStreamReader(content, head
                        .getName()));
            else
                reader = new BufferedReader(new InputStreamReader(content));

            while ((line = reader.readLine()) != null) {
                sb.append(line).append("\n");
            }
        } finally {
            content.close();
        }
        String str = sb.toString();
        return str;
    }

    protected void getSituationsByUrl(Date modifiedSince,
            ISituationHandler handler, String url) throws FetchInfoException {
        
        url += "&uid=" + mDeviceId;
        
        if (modifiedSince != null)
            url += "&modifiedSince=" + formatter.format(modifiedSince);
        
        try {
            JSONObject jSonResult = DownloadJSon(url);
            
            verifyApiVersion(jSonResult, url);
            
            JSONArray situationsJSonArray = jSonResult.getJSONArray("situations");
            
            int situationsCount = situationsJSonArray.length();
            for (int i = 0 ; i < situationsCount; i++){
               setStatus(Status.Extracting, i, situationsCount);
               JSONObject situation = situationsJSonArray.getJSONObject(i);
               TrafficInfo info = extractTrafficInfoFromJson(situation);
               handler.HandleSitation(info);
            }
            
            mLastTimestamp = extractDate(jSonResult, "timeStamp");
        } catch (JSONException e) {
            Log.e(LOGTAG, "JSONException in getSituations(2).");
            Log.e(LOGTAG, "JSONException message: " + e.getMessage());
            throw new FetchInfoException(url,
                    ErrorCodes.DataFormatError, e);
        }
    }

    @Override
    public Date getLastTimestamp() {
        return mLastTimestamp;
    }

    @Override
    public void SetStatusReciever(IStatusReciever reciver) {
        mStatusReciver = reciver;        
    }

    private void setStatus(Status status, int done, int total) {
        if (mStatusReciver != null)
            mStatusReciver.RecieveStatus(status, done, total);
    }
    
    private void setStatus(Status status){
        if (mStatusReciver != null)
            mStatusReciver.RecieveStatus(status);
    }
}
